@@ -21,6 +21,7 @@ errcode_t e2fsck_allocate_context(e2fsck_t *ret)
 {
 	e2fsck_t	context;
 	errcode_t	retval;
+	char		*time_env;
 
 	retval = ext2fs_get_mem(sizeof(struct e2fsck_struct), &context);
 	if (retval)
@@ -30,6 +31,13 @@ errcode_t e2fsck_allocate_context(e2fsck_t *ret)
 
 	context->process_inode_size = 256;
 	context->ext_attr_ver = 2;
+	context->blocks_per_page = 1;
+	
+	time_env = getenv("E2FSCK_TIME");
+	if (time_env)
+		context->now = strtoul(time_env, NULL, 0);
+	else
+		context->now = time(0);
 
 	*ret = context;
 	return 0;
@@ -158,6 +166,15 @@ void e2fsck_free_context(e2fsck_t ctx)
 	e2fsck_reset_context(ctx);
 	if (ctx->blkid)
 		blkid_put_cache(ctx->blkid);
+
+	if (ctx->profile)
+		profile_release(ctx->profile);
+
+	if (ctx->filesystem_name)
+		ext2fs_free_mem(&ctx->filesystem_name);
+			
+	if (ctx->device_name)
+		ext2fs_free_mem(&ctx->device_name);
 			
 	ext2fs_free_mem(&ctx);
 }
