@@ -55,6 +55,16 @@ extern int optind;
 #include "../version.h"
 #include "nls-enable.h"
 
+/* 
+ * Tune2fs supports these features in addition to the standard features.
+ */
+#define EXT2_TUNE2FS_INCOMPAT	(EXT3_FEATURE_INCOMPAT_EXTENTS)
+#define EXT2_TUNE2FS_RO_COMPAT	(EXT4_FEATURE_RO_COMPAT_HUGE_FILE|\
+				 EXT4_FEATURE_RO_COMPAT_GDT_CSUM|	\
+				 EXT4_FEATURE_RO_COMPAT_DIR_NLINK|	\
+				 EXT4_FEATURE_RO_COMPAT_EXTRA_ISIZE)
+
+
 const char * program_name = "tune2fs";
 char * device_name;
 char * new_label, *new_last_mounted, *new_UUID;
@@ -71,6 +81,9 @@ static unsigned short errors;
 static int open_flag;
 static char *features_cmd;
 static char *mntopts_cmd;
+static int stride, stripe_width;
+static int stride_set, stripe_width_set;
+static char *extended_cmd;
 
 int journal_size, journal_flags;
 char *journal_device;
@@ -84,21 +97,37 @@ static void usage(void)
 	fprintf(stderr,
 		_("Usage: %s [-c max_mounts_count] [-e errors_behavior] "
 		  "[-g group]\n"
-		  "\t[-i interval[d|m|w]] [-j] [-J journal_options]\n"
-		  "\t[-l] [-s sparse_flag] [-m reserved_blocks_percent]\n"
-		  "\t[-o [^]mount_options[,...]] [-r reserved_blocks_count]\n"
-		  "\t[-u user] [-C mount_count] [-L volume_label] "
-		  "[-M last_mounted_dir]\n"
-		  "\t[-O [^]feature[,...]] [-T last_check_time] [-U UUID]"
-		  " device\n"), program_name);
+		  "\t[-i interval[d|m|w]] [-j] [-J journal_options] [-l]\n"
+		  "\t[-m reserved_blocks_percent] "
+		  "[-o [^]mount_options[,...]] \n"
+		  "\t[-r reserved_blocks_count] [-u user] [-C mount_count] "
+		  "[-L volume_label]\n"
+		  "\t[-M last_mounted_dir] [-O [^]feature[,...]]\n"
+		  "\t[-E extended-option[,...]] [-T last_check_time] "
+		  "[-U UUID] device\n"), program_name);
 	exit (1);
 }
 
 static __u32 ok_features[3] = {
+	/* Compat */
 	EXT3_FEATURE_COMPAT_HAS_JOURNAL |
-		EXT2_FEATURE_COMPAT_DIR_INDEX,	/* Compat */
-	EXT2_FEATURE_INCOMPAT_FILETYPE,		/* Incompat */
-	EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER	/* R/O compat */
+		EXT2_FEATURE_COMPAT_DIR_INDEX,
+	/* Incompat */
+	EXT2_FEATURE_INCOMPAT_FILETYPE,
+	/* R/O compat */
+	EXT2_FEATURE_RO_COMPAT_LARGE_FILE |
+		EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER
+};
+
+static __u32 clear_ok_features[3] = {
+	/* Compat */
+	EXT3_FEATURE_COMPAT_HAS_JOURNAL |
+		EXT2_FEATURE_COMPAT_RESIZE_INODE |
+		EXT2_FEATURE_COMPAT_DIR_INDEX,
+	/* Incompat */
+	EXT2_FEATURE_INCOMPAT_FILETYPE,
+	/* R/O compat */
+	EXT2_FEATURE_RO_COMPAT_LARGE_FILE
 };
 
 /*
@@ -281,38 +310,44 @@ static void update_mntopts(ext2_filsys fs, char *mntopts)
  */
 static void update_feature_set(ext2_filsys fs, char *features)
 {
-	int sparse, old_sparse, filetype, old_filetype;
-	int journal, old_journal, dxdir, old_dxdir;
 	struct ext2_super_block *sb= fs->super;
 	__u32	old_compat, old_incompat, old_ro_compat;
+	__u32		old_features[3];
+	int		type_err;
+	unsigned int	mask_err;
 
-	old_compat = sb->s_feature_compat;
-	old_ro_compat = sb->s_feature_ro_compat;
-	old_incompat = sb->s_feature_incompat;
+#define FEATURE_ON(type, mask) (!(old_features[(type)] & (mask)) && \
+				((&sb->s_feature_compat)[(type)] & (mask)))
+#define FEATURE_OFF(type, mask) ((old_features[(type)] & (mask)) && \
+				 !((&sb->s_feature_compat)[(type)] & (mask)))
+#define FEATURE_CHANGED(type, mask) ((mask) & \
+		     (old_features[(type)] ^ (&sb->s_feature_compat)[(type)]))
 
-	old_sparse = sb->s_feature_ro_compat &
-		EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER;
-	old_filetype = sb->s_feature_incompat &
-		EXT2_FEATURE_INCOMPAT_FILETYPE;
-	old_journal = sb->s_feature_compat &
-		EXT3_FEATURE_COMPAT_HAS_JOURNAL;
-	old_dxdir = sb->s_feature_compat &
-		EXT2_FEATURE_COMPAT_DIR_INDEX;
-	if (e2p_edit_feature(features, &sb->s_feature_compat,
-			     ok_features)) {
-		fprintf(stderr, _("Invalid filesystem option set: %s\n"),
-			features);
+	old_features[E2P_FEATURE_COMPAT] = sb->s_feature_compat;
+	old_features[E2P_FEATURE_INCOMPAT] = sb->s_feature_incompat;
+	old_features[E2P_FEATURE_RO_INCOMPAT] = sb->s_feature_ro_compat;
+
+	if (e2p_edit_feature2(features, &sb->s_feature_compat,
+			      ok_features, clear_ok_features,
+			      &type_err, &mask_err)) {
+		if (!mask_err)
+			fprintf(stderr,
+				_("Invalid filesystem option set: %s\n"),
+				features);
+		else if (type_err & E2P_FEATURE_NEGATE_FLAG)
+			fprintf(stderr, _("Clearing filesystem feature '%s' "
+					  "not supported.\n"),
+				e2p_feature2string(type_err &
+						   E2P_FEATURE_TYPE_MASK,
+						   mask_err));
+		else
+			fprintf(stderr, _("Setting filesystem feature '%s' "
+					  "not supported.\n"),
+				e2p_feature2string(type_err, mask_err));
 		exit(1);
 	}
-	sparse = sb->s_feature_ro_compat &
-		EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER;
-	filetype = sb->s_feature_incompat &
-		EXT2_FEATURE_INCOMPAT_FILETYPE;
-	journal = sb->s_feature_compat &
-		EXT3_FEATURE_COMPAT_HAS_JOURNAL;
-	dxdir = sb->s_feature_compat &
-		EXT2_FEATURE_COMPAT_DIR_INDEX;
-	if (old_journal && !journal) {
+
+	if (FEATURE_OFF(E2P_FEATURE_COMPAT, EXT3_FEATURE_COMPAT_HAS_JOURNAL)) {
 		if ((mount_flags & EXT2_MF_MOUNTED) &&
 		    !(mount_flags & EXT2_MF_READONLY)) {
 			fputs(_("The has_journal flag may only be "
@@ -335,7 +370,8 @@ static void update_feature_set(ext2_filsys fs, char *features)
 			remove_journal_device(fs);
 		}
 	}
-	if (journal && !old_journal) {
+
+	if (FEATURE_ON(E2P_FEATURE_COMPAT, EXT3_FEATURE_COMPAT_HAS_JOURNAL)) {
 		/*
 		 * If adding a journal flag, let the create journal
 		 * code below handle creating setting the flag and
@@ -346,7 +382,8 @@ static void update_feature_set(ext2_filsys fs, char *features)
 			journal_size = -1;
 		sb->s_feature_compat &= ~EXT3_FEATURE_COMPAT_HAS_JOURNAL;
 	}
-	if (dxdir && !old_dxdir) {
+
+	if (FEATURE_ON(E2P_FEATURE_COMPAT, EXT2_FEATURE_COMPAT_DIR_INDEX)) {
 		if (!sb->s_def_hash_version)
 			sb->s_def_hash_version = EXT2_HASH_TEA;
 		if (uuid_is_null((unsigned char *) sb->s_hash_seed))
@@ -357,14 +394,22 @@ static void update_feature_set(ext2_filsys fs, char *features)
 	    (sb->s_feature_compat || sb->s_feature_ro_compat ||
 	     sb->s_feature_incompat))
 		ext2fs_update_dynamic_rev(fs);
-	if ((sparse != old_sparse) ||
-	    (filetype != old_filetype)) {
+
+	if (FEATURE_CHANGED(E2P_FEATURE_RO_INCOMPAT,
+			    EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER) ||
+	    FEATURE_CHANGED(E2P_FEATURE_INCOMPAT,
+			    EXT2_FEATURE_INCOMPAT_FILETYPE) ||
+	    FEATURE_CHANGED(E2P_FEATURE_COMPAT,
+			    EXT2_FEATURE_COMPAT_RESIZE_INODE) ||
+	    FEATURE_OFF(E2P_FEATURE_RO_INCOMPAT,
+			EXT2_FEATURE_RO_COMPAT_LARGE_FILE)) {
 		sb->s_state &= ~EXT2_VALID_FS;
 		printf("\n%s\n", _(please_fsck));
 	}
-	if ((old_compat != sb->s_feature_compat) ||
-	    (old_ro_compat != sb->s_feature_ro_compat) ||
-	    (old_incompat != sb->s_feature_incompat))
+
+	if ((old_features[E2P_FEATURE_COMPAT] != sb->s_feature_compat) ||
+	    (old_features[E2P_FEATURE_INCOMPAT] != sb->s_feature_incompat) ||
+	    (old_features[E2P_FEATURE_RO_INCOMPAT] != sb->s_feature_ro_compat))
 		ext2fs_mark_super_dirty(fs);
 }
 
@@ -460,8 +505,9 @@ static void parse_e2label_options(int argc, char ** argv)
 			argv[1]);
 		exit(1);
 	}
+	open_flag = EXT2_FLAG_SOFTSUPP_FEATURES | EXT2_FLAG_JOURNAL_DEV_OK;
 	if (argc == 3) {
-		open_flag = EXT2_FLAG_RW | EXT2_FLAG_JOURNAL_DEV_OK;
+		open_flag |= EXT2_FLAG_RW;
 		L_flag = 1;
 		new_label = argv[2];
 	} else 
@@ -494,6 +540,7 @@ static time_t parse_time(char *str)
 			str);
 		usage();
 	}
+	ts.tm_isdst = -1;
 	return (mktime(&ts));
 }
 
@@ -504,8 +551,10 @@ static void parse_tune2fs_options(int argc, char **argv)
 	struct group * gr;
 	struct passwd * pw;
 
+	open_flag = EXT2_FLAG_SOFTSUPP_FEATURES;
+
 	printf("tune2fs %s (%s)\n", E2FSPROGS_VERSION, E2FSPROGS_DATE);
-	while ((c = getopt(argc, argv, "c:e:fg:i:jlm:o:r:s:u:C:J:L:M:O:T:U:")) != EOF)
+	while ((c = getopt(argc, argv, "c:e:fg:i:jlm:o:r:s:u:C:E:J:L:M:O:T:U:")) != EOF)
 		switch (c)
 		{
 			case 'c':
@@ -548,6 +597,10 @@ static void parse_tune2fs_options(int argc, char **argv)
 				e_flag = 1;
 				open_flag = EXT2_FLAG_RW;
 				break;
+			case 'E':
+				extended_cmd = optarg;
+				open_flag |= EXT2_FLAG_RW;
+				break;
 			case 'f': /* Force */
 				f_flag = 1;
 				break;
@@ -618,7 +671,7 @@ static void parse_tune2fs_options(int argc, char **argv)
 			case 'L':
 				new_label = optarg;
 				L_flag = 1;
-				open_flag = EXT2_FLAG_RW |
+				open_flag |= EXT2_FLAG_RW |
 					EXT2_FLAG_JOURNAL_DEV_OK;
 				break;
 			case 'm':
@@ -667,7 +720,7 @@ static void parse_tune2fs_options(int argc, char **argv)
 				r_flag = 1;
 				open_flag = EXT2_FLAG_RW;
 				break;
-			case 's':
+			case 's': /* Deprecated */
 				s_flag = atoi(optarg);
 				open_flag = EXT2_FLAG_RW;
 				break;
@@ -739,6 +792,94 @@ void do_findfs(int argc, char **argv)
 	exit(0);
 }
 
+/*
+ * Note!  If any extended options are incompatible with the
+ * intersection of the SOFTSUPP features and those features explicitly
+ * enabled for tune2fs, there needs to be an explicit test for them
+ * here.
+ */
+static void parse_extended_opts(ext2_filsys fs, const char *opts)
+{
+	char	*buf, *token, *next, *p, *arg;
+	int	len;
+	int	r_usage = 0;
+
+	len = strlen(opts);
+	buf = malloc(len+1);
+	if (!buf) {
+		fprintf(stderr,
+			_("Couldn't allocate memory to parse options!\n"));
+		exit(1);
+	}
+	strcpy(buf, opts);
+	for (token = buf; token && *token; token = next) {
+		p = strchr(token, ',');
+		next = 0;
+		if (p) {
+			*p = 0;
+			next = p+1;
+		}
+		arg = strchr(token, '=');
+		if (arg) {
+			*arg = 0;
+			arg++;
+		}
+		if (!strcmp(token, "test_fs")) {
+			fs->super->s_flags |= EXT2_FLAGS_TEST_FILESYS;
+			printf("Setting test filesystem flag\n");
+			ext2fs_mark_super_dirty(fs);
+		} else if (!strcmp(token, "^test_fs")) {
+			fs->super->s_flags &= ~EXT2_FLAGS_TEST_FILESYS;
+			printf("Clearing test filesystem flag\n");
+			ext2fs_mark_super_dirty(fs);
+		} else if (strcmp(token, "stride") == 0) {
+			if (!arg) {
+				r_usage++;
+				continue;
+			}
+			stride = strtoul(arg, &p, 0);
+			if (*p || (stride == 0)) {
+				fprintf(stderr,
+				       _("Invalid RAID stride: %s\n"),
+					arg);
+				r_usage++;
+				continue;
+			}
+			stride_set = 1;
+		} else if (strcmp(token, "stripe-width") == 0 ||
+			   strcmp(token, "stripe_width") == 0) {
+			if (!arg) {
+				r_usage++;
+				continue;
+			}
+			stripe_width = strtoul(arg, &p, 0);
+			if (*p || (stripe_width == 0)) {
+				fprintf(stderr,
+					_("Invalid RAID stripe-width: %s\n"),
+					arg);
+				r_usage++;
+				continue;
+			}
+			stripe_width_set = 1;
+		} else 
+			r_usage++;
+	}
+	if (r_usage) {
+		fprintf(stderr, _("\nBad options specified.\n\n"
+			"Extended options are separated by commas, "
+			"and may take an argument which\n"
+			"\tis set off by an equals ('=') sign.\n\n"
+			"Valid extended options are:\n"
+			"\tstride=<RAID per-disk chunk size in blocks>\n"
+			"\tstripe-width=<RAID stride*data disks in blocks>\n"
+			"\ttest_fs\n"
+			"\t^test_fs\n"));
+		free(buf);
+		exit(1);
+	}
+	free(buf);
+}	
+
 
 int main (int argc, char ** argv)
 {
@@ -781,6 +922,13 @@ int main (int argc, char ** argv)
 	}
 	sb = fs->super;
 	fs->flags &= ~EXT2_FLAG_MASTER_SB_ONLY;
+	if ((sb->s_feature_incompat & !EXT2_TUNE2FS_INCOMPAT) ||
+	    (sb->s_feature_ro_compat & !EXT2_TUNE2FS_RO_COMPAT)) {
+		fprintf(stderr, 
+			_("Filesystem %s has unsupported features enabled.\n"),
+			device_name);
+		exit(1);
+	}
 	if (print_label) {
 		/* For e2label emulation */
 		printf("%.*s\n", (int) sizeof(sb->s_volume_name),
@@ -858,19 +1006,9 @@ int main (int argc, char ** argv)
 		}
 	}
 	if (s_flag == 0) {
-		if (!(sb->s_feature_ro_compat &
-		      EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER))
-			fputs(_("\nThe filesystem already has sparse "
-				"superblocks disabled.\n"), stderr);
-		else {
-			sb->s_feature_ro_compat &=
-				~EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER;
-			sb->s_state &= ~EXT2_VALID_FS;
-			fs->flags |= EXT2_FLAG_MASTER_SB_ONLY;
-			ext2fs_mark_super_dirty(fs);
-			printf(_("\nSparse superblock flag cleared.  %s"),
-			       _(please_fsck));
-		}
+		fputs(_("\nClearing the sparse superflag not supported.\n"),
+		      stderr);
+		exit(1);
 	}
 	if (T_flag) {
 		sb->s_lastcheck = last_check_time;
@@ -902,6 +1040,8 @@ int main (int argc, char ** argv)
 		update_mntopts(fs, mntopts_cmd);
 	if (features_cmd)
 		update_feature_set(fs, features_cmd);
+	if (extended_cmd)
+		parse_extended_opts(fs, extended_cmd);
 	if (journal_size || journal_device)
 		add_journal(fs);
 	
@@ -922,6 +1062,16 @@ int main (int argc, char ** argv)
 
 	if (l_flag)
 		list_super (sb);
+	if (stride_set) {
+		sb->s_raid_stride = stride;
+		ext2fs_mark_super_dirty(fs);
+		printf(_("Setting stride size to %d\n"), stride);
+	}
+	if (stripe_width_set) {
+		sb->s_raid_stripe_width = stripe_width;
+		ext2fs_mark_super_dirty(fs);
+		printf(_("Setting stripe width to %d\n"), stripe_width);
+	}
 	remove_error_table(&et_ext2_error_table);
 	return (ext2fs_close (fs) ? 1 : 0);
 }
