@@ -3,9 +3,9 @@
  * programs.
  * 
  * Copyright 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003,
- * 2004, 2005, 2006, 2007 by Theodore Ts'o.  This file may be
+ * 2004, 2005, 2006, 2007, 2008 by Theodore Ts'o.  This file may be
  * redistributed under the GNU Public License.
  */
 
-#define E2FSPROGS_VERSION "1.40.4"
-#define E2FSPROGS_DATE "31-Dec-2007"
+#define E2FSPROGS_VERSION "1.40.9"
+#define E2FSPROGS_DATE "27-Apr-2008"
